{$f+}
unit bigfile2;

(*  BIGFILES by V. Neri for Venom Software

  bigfile structure                   data type
  *****************                   *********

  filename (variable chars ) ____________     char
  data of file
  .
  filename (variable chars ) ____________     char
  data of file
  .
  filename (variable chars ) ____________     char
  data of file
  .
  .
  .
  --- header start (entries of 12 bytes) ---

  word checksum                         word
  filesize                              longint
  bytes from the end of the bigfile     longint
  .
  word checksum                         word
  filesize                              longint
  bytes from the end of the bigfile     longint
  .
  word checksum                         word
  filesize                              longint
  bytes from the end of the bigfile     longint
  .
  entries number                        word
  header start
  bytes from the end of the bigfile     longint
  --- header end ('.BGF') ---

*)
interface
{$f+}
function errorstring(error:byte):string;
function bigopen(name:string):byte;
function bigclose(bighandle:byte):byte;
function bigsearch(bighandle:byte;name:string):longint;
function openfile(bighandle:byte;name:string):word;
function closefile(bighandle:byte;handle:word):byte;
function readbig(bighandle:byte;handle:word;p:pointer;count:longint):byte;
function seekbig(bighandle:byte;handle:word;pos:longint):byte;
function bigsize(bighandle:byte):longint;
function sizefile(bighandle:byte;handle:word):longint;

{ Errors codes :
  255                   bigfile not found
  254                   file not found in bigfile
  253                   seek out of range
  251                   bad handle
  250                   write protection
  249                   file is not a bigfile
  248                   too many bigfiles
  247                   generic error
  245                   too many files
  0                     no error, all ok
  <240                  handle
}

implementation

const s='BIGFILE2 v2.0 - Valerio Neri';
filenamelength=50; { Length of Filename is variable }
var currentbig : byte;
    currentfile: word;
    dummy      : word;
    lastsize   : longint;
    longdummy  : longint;
    big : array[1..10] of record
                             filevar : file;
                             size    : longint;
                             handle  : byte;
                             active  : boolean;
                             header  : longint;
                             entries : word;
                          end;
    entry : record
               checksum : word;
               size     : longint;
               bytes    : longint;
            end;
    vfile : array[1..500] of longint;
 vfilesize: array[1..500] of longint;
  vfilepos: array[1..500] of longint;
{$f+}
function errorstring(error:byte):string;
begin
  case error of
    255   : errorstring:='Bigfile not found';
    254   : errorstring:='File not found in Bigfile';
    253   : errorstring:='Seek out of range';
    251   : errorstring:='Bad handle';
    250   : errorstring:='Write protection';
    249   : errorstring:='File is not a Bigfile';
    248   : errorstring:='Too many bigfiles open';
    247   : errorstring:='Internal error';
    245   : errorstring:='More than 500 files open';
      0   : errorstring:='Ok';
  else
  errorstring:='Not an Error. Handle number';
  end
end;
{$f+}
function bigopen(name:string):byte;
var checkstr : array[1..5] of char;
begin
  {$i-}
  currentbig:=1;
  while (big[currentbig].active=true)and(currentbig<11) do
    inc(currentbig);
  if currentbig=11 then
    begin
      bigopen:=248;
      exit;
    end;
  big[currentbig].handle:=currentbig;
  assign(big[currentbig].filevar,name);
  reset(big[currentbig].filevar,1);
  {$I+}
  if IOResult<>0 then
    begin
      bigopen:=255;
      exit;
    end;
  big[currentbig].size:=filesize(big[currentbig].filevar);
  { Check if .BGF available }
  {$i-}
  seek(big[currentbig].filevar,big[currentbig].size-4);
  blockread(big[currentbig].filevar,checkstr,5,dummy);
  if IOResult<>0 then
    begin
      bigopen:=247;
      close(big[currentbig].filevar);
      exit;
    end;
  if checkstr[1]<>'.' then
    begin
      bigopen:=249;
      close(big[currentbig].filevar);
      exit;
    end
  else
  if checkstr[2]<>'B' then
    begin
      bigopen:=249;
      close(big[currentbig].filevar);
      exit;
    end
  else
  if checkstr[3]<>'G' then
    begin
      bigopen:=249;
      close(big[currentbig].filevar);
      exit;
    end
  else
  if checkstr[4]<>'F' then
    begin
      bigopen:=249;
      close(big[currentbig].filevar);
      exit;
    end;
  { Ok, file is a bigfile }
  seek(big[currentbig].filevar,big[currentbig].size-10);
  blockread(big[currentbig].filevar,big[currentbig].entries,2,dummy);
  blockread(big[currentbig].filevar,longdummy,4,dummy);
  big[currentbig].header:=big[currentbig].size-longdummy-1;

  big[currentbig].active:=true;
  bigopen:=big[currentbig].handle;
end;

function sizefile(bighandle:byte;handle:word):longint;
begin
  if big[bighandle].active=false then
    sizefile:=0
  else
    if vfile[handle]=0
      then
        sizefile:=0
    else
      sizefile:=vfilesize[handle];
end;

function bigclose(bighandle:byte):byte;
begin
  if bighandle>10 then
    begin
      bigclose:=251;
      exit;
    end;
  {$i-}
  big[bighandle].active:=false;
  close(big[bighandle].filevar);
  if IOResult<>0 then
    begin
      bigclose:=247;
      exit;
    end;
end;

function bigsearch(bighandle:byte;name:string):longint;
var namecheck :word;
    gogo      :byte;
    counter   :word;
    entryname :array[1..filenamelength] of char;
    srcname   :array[1..filenamelength] of char;
    savepos   :longint;
begin
  for gogo:=1 to filenamelength do
    if gogo<=length(name) then
      srcname[gogo]:=upcase(name[gogo])
    else
      srcname[gogo]:=' ';

  if big[bighandle].active=false then
    begin
      bigsearch:=251;
      exit;
    end;
  namecheck:=0;
  for gogo:=1 to filenamelength do
    namecheck:=namecheck+ord(srcname[gogo]);
  seek(big[bighandle].filevar,big[bighandle].header);
  for counter:=big[bighandle].entries downto 1 do
  begin
    {$i-}
    blockread(big[bighandle].filevar,entry,10,dummy);
    savepos:=filepos(big[bighandle].filevar);
    if entry.checksum=namecheck then
      begin
        seek(big[bighandle].filevar,big[bighandle].size-entry.bytes-entry.size-filenamelength-1);
        blockread(big[bighandle].filevar,entryname,filenamelength,dummy);
        if IOResult<>0 then
          begin
            bigsearch:=247;
            exit;
          end;
        if entryname=srcname then
          begin
            lastsize:=entry.size;
            bigsearch:=filepos(big[bighandle].filevar);
            {bigsearch:=big[bighandle].size-entry.bytes+11;}
            exit;
          end;
      end;
    seek(big[bighandle].filevar,savepos);
  end;
  bigsearch:=0;
end;

function bigsize(bighandle:byte):longint;
begin
  bigsize:=big[bighandle].size;
  if big[bighandle].active=false then
    bigsize:=0;
end;

function openfile(bighandle:byte;name:string):word;
var position:longint;
begin
  currentfile:=1;
  position:=bigsearch(bighandle,name);
  if position=0 then
    begin
      openfile:=0;
      exit;
    end;
  while (currentfile<500)and(vfile[currentfile]<>0) do
      inc(currentfile);
  if currentfile=501 then
    begin
      openfile:=0;
      exit;
    end;
  vfile[currentfile]:=position;
  vfilesize[currentfile]:=lastsize;
  vfilepos[currentfile]:=0;
  openfile:=currentfile;
end;

function closefile(bighandle:byte;handle:word):byte;
begin
  if vfile[handle]=0 then
    begin
      closefile:=251;
      exit;
    end;
  vfile[handle]:=0;
  vfilesize[handle]:=0;
  closefile:=0;
end;

function readbig(bighandle:byte;handle:word;p:pointer;count:longint):byte;
var realposition:longint;
    realcount:longint;
begin
  if (vfile[handle]=0)or(handle=0) then
    begin
      readbig:=251;
      exit;
    end;
  realposition:=vfile[handle]+vfilepos[handle];
  if vfilepos[handle]+count>vfilesize[handle] then
    count:=abs(vfilepos[handle]-vfilesize[handle]);
  {$i-}
  seek(big[bighandle].filevar,realposition);
  blockread(big[bighandle].filevar,p^,count,realcount);
  if IOResult<>0 then
    begin
      readbig:=247;
      exit;
    end;
  inc(vfilepos[handle],realcount);
  readbig:=0;
end;

function seekbig(bighandle:byte;handle:word;pos:longint):byte;
begin
  if vfile[handle]=0 then
    begin
      seekbig:=251;
      exit;
    end;
  if pos>vfilesize[handle] then
    begin
      seekbig:=253;
      exit;
    end;
  inc(vfilepos[handle],pos);
  seekbig:=0;
end;

begin
  for dummy:=1 to 10 do
    with big[dummy] do
      active:=false;
  for dummy:=1 to 500 do
    vfile[dummy]:=0;
  for dummy:=1 to 500 do
    vfilesize[dummy]:=0;
  for dummy:=1 to 500 do
    vfilepos[dummy]:=0;
end.
